#ifndef bqtNewRadixSetHH
#define bqtNewRadixSetHH

#include "RadixTreeGMP-Vector.hh"
#include "NewRadixUtils.hh"

#include <bitset>

template<size_t RealSize>
class NewRadixSetGMP: private RadixTreeGMP<RealSize,0>
{
private:
    typedef RadixTreeGMP<RealSize,0> parent;
public:
    using parent::size;
    using parent::clear;
    inline bool find(BigInt& Key) const { return parent::find(Key); }
    inline bool add(BigInt& Key) { return parent::add(Key); }
};

/* Specializations for cases when the data fits in addressing space */

template<size_t RealSize>
class RadixSet_Bitset
{
private:
    static const size_t nbits = IntExp<8ULL*RealSize, 2>::Result;
    std::bitset<nbits> data;
    size_t count;
public:
    RadixSet_Bitset() : data(), count(0) { }
    virtual ~RadixSet_BitSet() { }
    
    inline size_t size() const { return count; /*data.count();*/ }
    void clear() { data.reset(); count=0; }
    
    inline bool find(const BigInt& Key) const
    {
        return data[ Key.GetInt() ];
    }
    inline bool add(const BigInt& Key)
    {
        size_t v = Key.GetInt(); 
        if(data[v]) return false;
        data.set(v); ++count;
        return true;
    }
};

template<> class NewRadixSetGMP<1>: public RadixSet_Bitset<1> { };
template<> class NewRadixSetGMP<2>: public RadixSet_Bitset<2> { };
template<> class NewRadixSetGMP<3>: public RadixSet_Bitset<3> { };
#ifdef __x86_64
template<> class NewRadixSetGMP<4>: public RadixSet_Bitset<4> { };
template<> class NewRadixSetGMP<5>: public RadixSet_Bitset<5> { };
//template<> class NewRadixSetGMP<6>: public RadixSet_Bitset<6> { };
#endif


/* Note: BytesUsedBy won't work if MaxValue-MinValue == (size_t)~0 */
template<typename KeyType, size_t MaxValue, size_t KeyLength, size_t MinValue=0>
class NewRadixSet:
    private NewRadixSetGMP<BytesUsedBy<MaxValue-MinValue+1,KeyLength>::MaxBytes>
{
private:
    typedef NewRadixSetGMP<BytesUsedBy<MaxValue-MinValue+1,KeyLength>::MaxBytes> parent;
    static inline BigInt& Encode(const KeyType*const Key)
    {
        return NewRadixEncoder::Encode<KeyType,MaxValue,MinValue> (Key, KeyLength);
    }
public:
    using parent::size;
    using parent::clear;
    inline bool find(const KeyType*const Key) const { return parent::find(Encode(Key)); }
    inline bool add(const KeyType*const Key) { return parent::add(Encode(Key)); }
};

#endif
