//#define PAIR_ATTRIBUTES __attribute__((packed))

#define _LARGEFILE64_SOURCE
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>

#include <algorithm>
#include <climits>
#include <iostream>
#include <sstream>

#include <csignal>

#ifdef __CYGWIN__
# define off64_t _off64_t
# define O_LARGEFILE 0
# define lseek64 lseek
# define mmap64 mmap
#endif

#include "levelstate.hh"
#include "deadzones.hh"
#include "completer.hh"

static void CompleteLevel()
{
    DeadZones.BuildFromLevel();
    
    if(TerminateSignalled)
    {
        std::cerr << "Terminate signalled, not solving a level!\n";
        return;
    }

    LevelState state;
    state.CreateFromLevel();
    state.RebuildBlockCaches();
    
    Completer completer(state);
    
    PlayingNormally=true;
//    ShowDead=true;    
    while(!TerminateSignalled)
    {
        bool Finished = completer.Iteration();
        if(Finished) break;
    }
    PlayingNormally=false;
    completer.DumpBest();
    
    std::cout << completer.GetJs();
}

static void CompleteLevel(int phase,int level)
{
    levelno = (phase-1)*5+level-1;
    std::cout << "/* LEVEL " << levelno << " */\n";
    CompleteLevel();
}

static void Terminate(int sig)
{
    TerminateSignalled=true;
}

Completer::active_list_type Completer::active_list;

int main(int argc, const char* const* argv)
{
    signal(SIGTERM, Terminate);
    signal(SIGINT,  Terminate);
    
#if CHOOSE_LEVEL == 60
    for(levelno=0; levelno<60; ++levelno)
    {
        DeadZones.BuildFromLevel();
        std::cerr << std::flush;
    }
    return 0;
#endif
#if 0
    LevelState state;
    state.CreateFromLevel();
    state.RebuildBlockCaches();
    state.Walk(-1,-2);
    state.Kick(-1,0,1);
    state.Walk(0,-2);
    state.CanPush(-1,0);
    state.Kick(-1,0,1);
    state.Walk(-3,3);
    //state.CanPush(0,-1);
    return 0;
#endif
    
    if(argc < 3)
    {
        levelno=CHOOSE_LEVEL;
        CompleteLevel();
        return 0;
    }
    
    CompleteLevel(std::atoi(argv[1]), std::atoi(argv[2]));
    return 0;
}
