#include "binom.hh"
#include "deadzones.hh"
#include "levelstate.hh"

unsigned DeadCounter1=0, DeadCounter0=0;
unsigned MultiDeadCounter[USE_MULTI_DEAD+1]={0};
bool ShowDead=false;
bool PlayingNormally=false;

ObstacleMapType DefaultObstacleMap;
ObstacleMapType CurrentObstacleMap;
std::bitset<FloorAmount> CurrentBlockMap;

char WallMap[MaxDim]; /* Array of wall and goal bits */

BigInt& LevelIdentifier::GetCompactData() const
{
    static BigInt compact;

    compact = 0;

    const unsigned MaxCoord = CompactBlockCoordinateLimit;
    const unsigned MaxBox   = MaxObject-1;
    for(unsigned Last=0, nbox=0; nbox<MaxBox; ++nbox)
    {
        unsigned Coord = CompactBlockCoordinateTranslation[ID[nbox+1]];
        compact += Binom::Choose(MaxCoord-Last,  MaxBox-nbox);
        compact -= Binom::Choose(MaxCoord-Coord, MaxBox-nbox);
        Last=Coord+1;
    }
    
    /* Since a player can't be in the same position as blocks are,
     * eliminate the block positions from the formula.
     */
    const CoordinateType* b = ID+1, *e = ID+MaxObject;
    unsigned minus = std::upper_bound(b, e, ID[0]) - b;
    
    compact *= CompactPlayerCoordinateLimit;
    compact += CompactPlayerCoordinateTranslation[ID[0]] - minus;
    
    return compact;
}

void LevelStateData::BuildFromLevel()
{
    const mapdata& map = maps[levelno];
    
    unsigned px = map.px - OffsetWidth;
    unsigned py = map.py - OffsetHeight;
    
    State.Reset();
    State.SetPlayer(py*MaxWidth+px);
    
    DefaultObstacleMap.reset();
    
    for(unsigned y=0; y<MaxHeight; ++y)
        for(unsigned x=0; x<MaxWidth; ++x)
        {
            unsigned c = map.mapdata[(y+OffsetHeight)*19+(x+OffsetWidth)] & 7;
            if(c & MapBits::Block)
            {
                State.RedefineBlock(NonCoordinate, y*MaxWidth+x);
            }
            
            WallMap[y*MaxWidth + x] = (c & 3);
            if(c & MapBits::Wall)
            {
                DefaultObstacleMap.set(y*MaxWidth + x);
            }
        }
}


void LevelStateData::Dump(unsigned focus) const
{
    bool FocusFound=false;
    bool FoundNonObstacleBlock=false;
    unsigned nblocks=0;
    for(unsigned y=0; y<MaxHeight; ++y)
    {
        for(unsigned x=0; x<MaxWidth; ++x)
        {
            unsigned pos = y*MaxWidth+x;
            
            //if(CurrentObstacleMap.test(pos)) std::cerr << "[44m";
            
            unsigned floorpos = FloorCoordinateTranslation[pos];
            if(floorpos < FloorAmount
            && CurrentBlockMap.test(floorpos))
            {
                std::cerr << "[1;31m";
                if(!CurrentObstacleMap.test(pos)) FoundNonObstacleBlock=true;
                nblocks += 1;
            }
            if(pos==focus) std::cerr << "[5m";
            
            if(IsWall(pos))
                std::cerr << '#';
            else if(State.HasBlock(pos))
            {
                if(pos==focus) { FocusFound=true; }
                std::cerr << '@';
            }
            else if(pos == State.GetPlayer())
                std::cerr << 'P';
            else if(IsGoal(pos))
                std::cerr << "[32m*";
            else if(DeadZones.CheckDead(pos))
                std::cerr << "[34md";
            else
                std::cerr << ' ';
            
            std::cerr << "[0m";
        }
        std::cerr << "\n";
    }
    
    
    if(focus != NonCoordinate && !FocusFound)
    {
        std::cerr << "ERROR: Focus " << focus << " not found!\n";
    }
    if(FoundNonObstacleBlock) std::cerr << "ERROR: Found non-obstacle block!\n";
    if(PlayingNormally && 
       nblocks != MaxObject-1) std::cerr << "ERROR: Block amount doesn't match!\n";
    
    std::cerr << "\33[2A\33[" << (MaxWidth+4) << "C" << std::flush;
    State.Dump();
    
    std::cerr << "\r\33[" << (MaxWidth+4) << "C" << std::flush;

    std::fprintf(stderr, "Compacted:");
    BigInt& key_enc = State.GetCompactData();       
    BigInt rem_enc;
    enum { RealSize = MaxCompactCoordinateData };
    unsigned char KeySeq[RealSize];   
    for(unsigned a=0; a<RealSize; ++a)
    {
        unsigned remainder = key_enc.DivMod2Exp(key_enc, rem_enc, 8);
        KeySeq[RealSize-a-1] = remainder;
    }
    for(unsigned a=0; a<RealSize; ++a) std::fprintf(stderr, " %02X", KeySeq[a]);
    std::fprintf(stderr, "\n");
    std::fflush(stderr);
}


bool LevelState::CheckBlunder(unsigned blockpos, int offset) const /*REGPARM*/
{
    //if(!PlayingNormally)return false;
    
    unsigned plrpos = blockpos-offset;

    /* blockpos  = new location of the block */
    if(DeadZones.CheckDead(blockpos))
    {
        /*
        if(ShowDead)
        {
            std::cerr << "--- Blunder 1 ? ---\n";
            unsigned tmp = GetPlayerPos();
            (const_cast<LevelState*>(this))->WalkTo(plrpos);
            Dump(blockpos);
            (const_cast<LevelState*>(this))->WalkTo(tmp);
        }
        */

        ++DeadCounter1;
        return true;
    }

    if(DeadZones.CheckMultiBlockDead(plrpos, blockpos))
    {
        if(ShowDead)
        {
            std::cerr << "--- Multiblunder ---\n";
            unsigned tmp = GetPlayerPos();
            (const_cast<LevelState*>(this))->WalkTo(plrpos);
            Dump(blockpos);
            (const_cast<LevelState*>(this))->WalkTo(tmp);
        }
        return true;
    }
    
#if USE_MULTI_DEAD >= 4
    if(PlayingNormally) return false;
#endif

    //SimpleVec<int,MaxObject> Movable;
    int Movable[MaxObject];
    Movable[0] = 2;
    
    //SimpleVec<unsigned,MaxDim> Identify;
    unsigned char Identify[MaxDim] = { 0 };

    CoordinateType BlockPos[MaxObject];
    for(unsigned n1=1; n1<MaxObject; ++n1)
    {
        unsigned n1pos = GetBlockPos(n1-1);
        BlockPos[n1] = n1pos;
        if(unlikely(n1pos == NonCoordinate)) break;

        //if(n1pos >= MaxDim) { std::cerr << n1pos << std::endl; }
        Identify[n1pos] = n1;
        
        if(PlayingNormally)
        {
            /*
            Optimize:
              - If the block is at goal, it's "maybe"
              - If the block is the given block, it's "maybe"
              - If the block is adjacent to given block, it's "maybe"
              - Otherwise it's "movable=yes", because assume we've not done blunder yet.
              
              - Diagonals are also included in the adjacency test.
                Otherwise the Magic Four is not detected as a blunder.
            */
            if((n1pos < blockpos         -1 || n1pos > blockpos         +1)
            && (n1pos < blockpos-MaxWidth-1 || n1pos > blockpos-MaxWidth+1)
            && (n1pos < blockpos+MaxWidth-1 || n1pos > blockpos+MaxWidth+1)
            && !data.IsGoal(n1pos))
            {
                Movable[n1] = 2;
            }
            else
                Movable[n1] = 1;
        }
        else
        {
            /* Not playing normally, thus we can't assume
             * that a blunder hasn't been made
             */
            Movable[n1] = 1;
        }
    }/*for n1*/

    
    #define Movability_1dir(frontpos) \
        (data.IsWall(frontpos) ? 0 : Movable[Identify[frontpos]]) \
    /* If there is no obstacle in that direction, it's movable (objno=0, movable=Yes) */
    
    /* No bounds checking! The field should have walls! */
    #define Movability_4dirs(blockpos) \
        ({ unsigned m1 = Movability_1dir(blockpos+1); \
           unsigned m2 = Movability_1dir(blockpos-1); \
           unsigned m3 = Movability_1dir(blockpos+MaxWidth); \
           unsigned m4 = Movability_1dir(blockpos-MaxWidth); \
           ((m1==0||m2==0) && (m3==0||m4==0)) ? 0 : \
           ((m1==2&&m2==2) || (m3==2&&m4==2)) ? 2 : \
           1; \
        })
    
    unsigned rolls;
    for(rolls=1; ; ++rolls)
    {
        bool missing=false, helped=false;
        for(unsigned n=1; n<MaxObject; ++n)
        {
            unsigned npos = BlockPos[n];
            if(unlikely(npos == NonCoordinate)) break;
            
            int CurrentState = Movable[n];
            if(CurrentState != 1)
            {
                continue; // This block already determined
            }
            
            int NewState = Movability_4dirs(npos);
            if(NewState != CurrentState)
            {
                helped=true;
                Movable[n] = NewState;
            }
            if(NewState == 1)
                missing = true;
        }
        
        /* If nothing to solve, or nothing was solved, break */
        if(!missing || !helped)
        {
            break;
        }
    }
    
    for(unsigned n=1; n<MaxObject; ++n)
    {
        unsigned npos = BlockPos[n];
        if(unlikely(npos == NonCoordinate)) break;
    
        int State = Movable[n];
        /* If the block could not be proven to be moveable,
         * it must be stuck.
         */
        if(State < 2)
        {
            /* If the block is not movable, it must be in a goal or it's blunder */
            if(!data.IsGoal(npos))
            {
             /*   std::cerr << "Blunder, in " << rolls << " rounds:\n";
                Dump(npos);*/
                ++DeadCounter0;
                return true;
            }
            else
            {/*
                std::cerr << "Unmovable, but not blunder:\n";
                Dump(npos);*/
            }
        }
    }
    return false;
}

void LevelState::FindWalkingDestinations(FavorablePointListWithCost& targets) const
{
    ObstacleMapType UnwantedPoints = CurrentObstacleMap;

    /* Find all positions that could be favorable */
    /* Such positions are next to each block. */
    std::bitset<MaxDim> Favorable;
    
    for(unsigned n=1; n<MaxObject; ++n)
    {
        unsigned pos = GetBlockPos(n-1); // location of a block
        if(unlikely(pos == NonCoordinate)) break;
        /* No bounds checking! The field should have walls! */
        /* It doesn't matter if the favorable points are inside walls,
         * but we check it anyway in hopes that the !any() condition
         * will match more often.
         */
        #define Try(p) { if(!UnwantedPoints.test(p)) Favorable.set(p); }
        Try(pos-1);
        Try(pos+1);
        Try(pos-MaxWidth);
        Try(pos+MaxWidth);
        #undef Try
    }

    SimpleVec<CoordinateType, MaxDim> CurPoints, NextPoints;
    const unsigned origin = data.State.GetPlayer();
    CurPoints.push_back(origin);
    Favorable.set(origin); // Consider the origin favorable too
    UnwantedPoints.set(origin); // But the origin should never be returned into

    for(unsigned Cost=0; !CurPoints.empty(); ++Cost)
    {
        NextPoints.clear();
        for(unsigned a=CurPoints.size(); a-->0; )
        {
            unsigned curpoint = CurPoints[a];
            if(Favorable.test(curpoint))
            {
                Favorable.reset(curpoint);
                targets.push_back(FavorablePointWithCost(curpoint, Cost));
                if(unlikely(!Favorable.any())) return;
            }

            #define Try(p) { unsigned newpoint = p; \
                if(!UnwantedPoints.test(newpoint)) \
                { \
                    UnwantedPoints.set(newpoint); \
                    NextPoints.push_back(newpoint); \
                } }

            /* No bounds checking! The field should have walls! */
            Try(curpoint-MaxWidth);
            Try(curpoint-1);
            Try(curpoint+1);
            Try(curpoint+MaxWidth);
            
            #undef Try
        }
        CurPoints.quickswap(NextPoints);
    }
}

void LevelState::FindWalkingDestinations(FavorablePointList& targets) const
{
    ObstacleMapType UnwantedPoints = CurrentObstacleMap;

    /* Find all positions that could be favorable */
    /* Such positions are next to each block. */
    std::bitset<MaxDim> Favorable;
    
    for(unsigned n=1; n<MaxObject; ++n)
    {
        unsigned pos = GetBlockPos(n-1); // location of a block
        if(unlikely(pos == NonCoordinate)) break;
        /* No bounds checking! The field should have walls! */
        /* It doesn't matter if the favorable points are inside walls,
         * but we check it anyway in hopes that the !any() condition
         * will match more often.
         */
        #define Try(p) { if(!UnwantedPoints.test(p)) Favorable.set(p); }
        Try(pos-1);
        Try(pos+1);
        Try(pos-MaxWidth);
        Try(pos+MaxWidth);
        #undef Try
    }

    SimpleVec<CoordinateType, MaxDim> CurPoints, NextPoints;
    const unsigned origin = data.State.GetPlayer();
    CurPoints.push_back(origin);
    Favorable.set(origin); // Consider the origin favorable too
    UnwantedPoints.set(origin); // But the origin should never be returned into

    while(!CurPoints.empty())
    {
        NextPoints.clear();
        for(unsigned a=CurPoints.size(); a-->0; )
        {
            unsigned curpoint = CurPoints[a];
            if(Favorable.test(curpoint))
            {
                Favorable.reset(curpoint);
                targets.push_back(curpoint);
                if(unlikely(!Favorable.any())) return;
            }

            #define Try(p) { unsigned newpoint = p; \
                if(!UnwantedPoints.test(newpoint)) \
                { \
                    UnwantedPoints.set(newpoint); \
                    NextPoints.push_back(newpoint); \
                } }

            /* No bounds checking! The field should have walls! */
            Try(curpoint-MaxWidth);
            Try(curpoint-1);
            Try(curpoint+1);
            Try(curpoint+MaxWidth);
            
            #undef Try
        }
        CurPoints.quickswap(NextPoints);
    }
}

void LevelState::FindWalkingRouteTo(unsigned nx,unsigned ny, WalkingRoute& result) const
{
    WalkingRoute Routes[MaxDim];
    
    ObstacleMapType UnwantedPoints = CurrentObstacleMap;
    
    #define Reassign(pos, route, newpoint) \
        Routes[pos] = route; Routes[pos].push_back(newpoint);
    
    SimpleVec<CoordinateType, MaxDim> CurPoints, NextPoints;
    const unsigned origin = data.State.GetPlayer();
    const unsigned goal   = ny*MaxWidth+nx;
    CurPoints.push_back(origin);
    
    for(unsigned Cost=0; !CurPoints.empty(); ++Cost)
    {
        NextPoints.clear();
        for(unsigned a=CurPoints.size(); a-->0; )
        {
            unsigned curpoint = CurPoints[a];
            const WalkingRoute& curroute = Routes[curpoint];
            
            if(curpoint == goal)
            {
                result = curroute;
                return;
            }

            #define Try(dir) { unsigned newpoint=curpoint+dir; \
                if(!UnwantedPoints.test(newpoint)) \
                { \
                    UnwantedPoints.set(newpoint); \
                    Reassign(newpoint, curroute, newpoint); \
                    NextPoints.push_back(newpoint); \
                } }

            /* No bounds checking! The field should have walls! */
            Try(-MaxWidth);
            Try(-1);
            Try(1);
            Try(MaxWidth);
            
            #undef Try
        }
        CurPoints.quickswap(NextPoints);
    }
    #undef Reassign
    
    result = Routes[goal];
}

bool LevelState::TestWalkingRoute(const unsigned origin, const unsigned goal) const
{
    ObstacleMapType UnwantedPoints = CurrentObstacleMap;
    
    SimpleVec<CoordinateType, MaxDim> CurPoints, NextPoints;
    CurPoints.push_back(origin);
    UnwantedPoints.set(origin);
    
    while(!CurPoints.empty())
    {
        NextPoints.clear();
        for(unsigned a=CurPoints.size(); a-->0; )
        {
            unsigned curpoint = CurPoints[a];
            if(curpoint == goal) return true;

            #define Try(dir) { unsigned newpoint=curpoint+dir; \
                if(!UnwantedPoints.test(newpoint)) \
                { \
                    UnwantedPoints.set(newpoint); \
                    NextPoints.push_back(newpoint); \
                } }

            /* No bounds checking! The field should have walls! */
            Try(-MaxWidth);
            Try(-1);
            Try(1);
            Try(MaxWidth);
            
            #undef Try
        }
        CurPoints.quickswap(NextPoints);
    }
    return false;
}

SimpleVec<CoordinateType,MaxDim> LevelState::BuildWalkingMaps() const
{
    SimpleVec<CoordinateType,MaxDim> result;
    result.resize(MaxDim);
    std::fill(result.begin(), result.end(), 0);
    
    ObstacleMapType UnwantedPoints = CurrentObstacleMap;
    
    SimpleVec<CoordinateType, MaxDim> CurPoints, NextPoints;
    for(unsigned plrpos=0; plrpos<MaxDim; ++plrpos)
    {
        if(UnwantedPoints.test(plrpos)) continue;
        
        CurPoints.clear();
        CurPoints.push_back(plrpos);
        UnwantedPoints.set(plrpos);
        
        while(!CurPoints.empty())
        {
            NextPoints.clear();
            for(unsigned a=CurPoints.size(); a-->0; )
            {
                unsigned curpoint = CurPoints[a];
                result[curpoint] = plrpos;

                #define Try(dir) { unsigned newpoint=curpoint+dir; \
                    if(!UnwantedPoints.test(newpoint)) \
                    { \
                        UnwantedPoints.set(newpoint); \
                        NextPoints.push_back(newpoint); \
                    } }

                /* No bounds checking! The field should have walls! */
                Try(-MaxWidth);
                Try(-1);
                Try(1);
                Try(MaxWidth);
                
                #undef Try
            }
            CurPoints.quickswap(NextPoints);
        }
    }
    return result;
}
