#include "settings.hh"
#include "templates.hh"

#include <iostream>
#include <cstdlib>

//#define TIMESTAMP_MOVES

struct Move
{
    enum types { WalkToKick, WalkToPush };
    enum dirs  { Up,Left,Right,Down     };
    
#ifdef TIMESTAMP_MOVES
    unsigned timestamp : 16;
#endif
    CoordinateType  pos:  Log2Ceil<MaxDim>::Bits;
    unsigned   type: Log2Ceil<2>::Bits; // 1
    unsigned   dir:  Log2Ceil<4>::Bits; // 2
    unsigned offsetbytes: Log2Ceil<4>::Bits; // 2
    
    /* Unfortunately g++ doesn't allow zero-width bitfields. */
    
    enum { offset_extra_bits = 8-((Log2Ceil<MaxDim>::Bits+5)&7) };
    unsigned offset_extra: offset_extra_bits;

public:
    Move(types t, CoordinateType where, int facing) : pos(where),type(t)
    {
        SetFacing(facing);
    }
    
    inline int getdx() const
    {
        static const signed char MoveDirXtable[4] = {0,-1,1,0};
        return MoveDirXtable[dir];
    }
    inline int getdy() const
    {
        static const signed char MoveDirYtable[4] = {-1,0,0,1};
        return MoveDirYtable[dir];
    }
    inline unsigned getpx() const { return pos%MaxWidth; }
    inline unsigned getpy() const { return pos/MaxWidth; }
    
    inline int GetFacing() const
    {
        static const int MoveDirTable[4] = {-MaxWidth,-1,1,MaxWidth};
        return MoveDirTable[dir];
    }
    void SetFacing(int facing)
    {
        switch(facing)
        {
            case -1: dir=Left; break;
            case  1: dir=Right; break;
            case -MaxWidth: dir=Up; break;
            case  MaxWidth: dir=Down; break;
            default:
                abort();
        }
    }
    
    void Dump(std::ostream& out = std::cout) const;
}   __attribute__((packed));

class MoveTree
{
private:
    unsigned Last;
    static const unsigned Root = (unsigned)(~0);
public:
    MoveTree() : Last(Root)
    {
    }
    ~MoveTree()
    {
        clear();
    }
    MoveTree(const MoveTree& b): Last(b.Last)
    {
    }
    MoveTree(const MoveTree& b, const Move& t) : Last(b.Last)
    {
        push_back(t);
    }

    bool empty() const { return rbegin() == rend(); }
    const Move& back() const { return *rbegin(); }

    void push_back(const Move& t);
    void clear();
    unsigned size(unsigned maxsize=~0) const
    {
        unsigned Size = 0;
        for(reverse_const_iterator i = rbegin(); i != rend(); ++i)
        {
            if(++Size >= maxsize) break;
        }
        return Size;
    }
    
    const Move& operator[] (unsigned n) const
    {
        n = size()-1-n;
        reverse_const_iterator i = rbegin();
        while(n > 0) { ++i; --n; }
        return *i;
    }
    
public:
    struct reverse_const_iterator
    {
        inline reverse_const_iterator(unsigned p): ptr(p) { }
        inline bool operator== (const reverse_const_iterator& b) const { return ptr==b.ptr; }
        inline bool operator!= (const reverse_const_iterator& b) const { return ptr!=b.ptr; }

        reverse_const_iterator& operator++ ();
        const Move& operator* ();
    private:
        unsigned ptr;
    };
    reverse_const_iterator rbegin() const { return reverse_const_iterator(Last); }
    reverse_const_iterator rend() const { return reverse_const_iterator(Root); }
} __attribute__((packed));
