#ifndef bqtTemplatesHH
#define bqtTemplatesHH

#define likely(x)   __builtin_expect(!!(x), 1)
#define unlikely(x) __builtin_expect(!!(x), 0)

template <int T>
struct BinaryComponentSize
{
    enum { _Divisible = !(T&1),
           Result = _Divisible ? BinaryComponentSize<T/2>::Result+1 : 0 };
};
template <> struct BinaryComponentSize<0> { enum { Result=0 }; };
template <> struct BinaryComponentSize<1> { enum { Result=0 }; };

template<unsigned p,unsigned v> struct IntExp
    { static const unsigned Result = v * IntExp<p-1,v>::Result; };
template<unsigned v> struct IntExp<1,v> { static const unsigned Result = v; };
template<unsigned v> struct IntExp<0,v> { static const unsigned Result = 1; };

template<unsigned T> struct Log2Floor { enum { Bits = Log2Floor<T/2>::Bits + 1 }; };
template<> struct Log2Floor<1> { enum { Bits = 0 }; };
template<> struct Log2Floor<0> { enum { Bits = -1 }; };
template<unsigned T> struct Log2Ceil
{
    enum { PowerOf2 = !(T&(T-1)), Bits = Log2Floor<T/2>::Bits + 2 - PowerOf2 };
};

template <unsigned T, unsigned Exp>
struct Powvalue
{
    enum { _Binary = BinaryComponentSize<T>::Result,
           Shift = _Binary*Exp };
    static const unsigned Mul = IntExp<Exp, T >> _Binary>::Result;
};

template<int T> struct OptimalTypePerRemainder { typedef unsigned char Type; };

template<> struct OptimalTypePerRemainder<sizeof(long)%sizeof(long)>
    { typedef unsigned long Type; };
template<> struct OptimalTypePerRemainder<sizeof(short)%sizeof(long)>
    { typedef unsigned short Type; };
template<int Nbytes> struct OptimalStorageType
    { typedef typename OptimalTypePerRemainder<Nbytes%sizeof(long)>::Type Type; };

template<int T> struct CoordTraits { };
template<> struct CoordTraits<1> { typedef unsigned char  Type; };
template<> struct CoordTraits<2> { typedef unsigned short Type; };
template<> struct CoordTraits<3> { typedef unsigned int Type; };
template<> struct CoordTraits<4> { typedef unsigned int Type; };

template<unsigned T>
struct Log256Ceil
{
    enum { Result = (Log2Ceil<T>::Bits + 7) / 8 };
};

template<unsigned p> struct Max2Exp
{
    enum { Result = (sizeof(unsigned)*8) / Log2Ceil<p>::Bits };
};
template<unsigned p> struct Max256Exp
{
    enum { Result = (sizeof(unsigned)*8) / Log256Ceil<p>::Bits };
};

template<unsigned OuterLimit, unsigned MaxCount>
struct BytesUsedBy
{
    enum { MaxBits = 
      (MaxCount * (Log2Ceil<
                    IntExp<Max2Exp<OuterLimit>::Result, OuterLimit>::Result
                        >::Bits) / Max2Exp<OuterLimit>::Result) };
    enum { MaxBytes = (MaxBits+7)/8 };
};

#endif
