#ifndef bqtNewRadixSetHH
#define bqtNewRadixSetHH

#include "RadixTreeGMP-Vector.hh"
#include "NewRadixUtils.hh"

template<unsigned RealSize>
class NewRadixSetGMP: private RadixTreeGMP<RealSize,0>
{
private:
    typedef RadixTreeGMP<RealSize,0> parent;
public:
    using parent::size;
    using parent::clear;
    inline bool find(BigInt& Key) const { return parent::find(Key); }
    inline bool add(BigInt& Key) { return parent::add(Key); }
};

/* Note: BytesUsedBy won't work if MaxValue-MinValue == (unsigned)~0 */
template<typename KeyType, unsigned MaxValue, unsigned KeyLength, unsigned MinValue=0>
class NewRadixSet:
    private NewRadixSetGMP<BytesUsedBy<MaxValue-MinValue+1,KeyLength>::MaxBytes>
{
private:
    typedef NewRadixSetGMP<BytesUsedBy<MaxValue-MinValue+1,KeyLength>::MaxBytes> parent;
    static inline BigInt& Encode(const KeyType*const Key)
    {
        return NewRadixEncoder::Encode<KeyType,MaxValue,MinValue> (Key, KeyLength);
    }
public:
    using parent::size;
    using parent::clear;
    inline bool find(const KeyType*const Key) const { return parent::find(Encode(Key)); }
    inline bool add(const KeyType*const Key) { return parent::add(Encode(Key)); }
};

#endif
