#include "coordinates.hh"

#include "levelstate.hh"
#include "deadzones.hh"
#include "moves.hh"

#include <algorithm>
#include <iostream>
#include <cmath> /* For std::log */

volatile bool TerminateSignalled = false;
unsigned levelno = CHOOSE_LEVEL;
unsigned MaxDuration = InitialMaxDuration;
unsigned ActualMaxObject=0;

unsigned CompactPlayerCoordinateTranslation[MaxDim];
unsigned CompactPlayerCoordinateLimit;
unsigned CompactBlockCoordinateTranslation[MaxDim+1];
unsigned CompactBlockCoordinateLimit;
unsigned FloorCoordinateTranslation[MaxDim];
unsigned FloorCoordinateUntranslation[FloorAmount];

#if CHOOSE_LEVEL == 60
static unsigned GetMaxDuration()
{
    unsigned result = 60*30*10; // 10 minutes
    
    std::string f1,f2;
    
    {std::stringstream s;
    s << "../solution/" << levelno << ".txt";
    f1 = s.str();}
    
    {std::stringstream s;
    s << "../records/" << levelno << ".txt";
    f2 = s.str();}
    
    fprintf(stderr, "/* completion time of %s and %s ",
        f1.c_str(),f2.c_str()); fflush(stderr);
    
    {FILE*fp=fopen(f1.c_str(),"rb");if(!fp)f1="";else{fseek(fp,0,SEEK_END);
     std::vector<unsigned char> buf(ftell(fp));rewind(fp);
     fread(&buf[0],1,buf.size(),fp);fclose(fp); f1.assign(buf.begin(),buf.end()); }}

    {FILE*fp=fopen(f2.c_str(),"rb");if(!fp)f2="";else{fseek(fp,0,SEEK_END);
     std::vector<unsigned char> buf(ftell(fp));rewind(fp);
     fread(&buf[0],1,buf.size(),fp);fclose(fp); f2.assign(buf.begin(),buf.end()); }}
    
    unsigned p = f1.find("perfect_data=");
    if(p!=f1.npos)
    {
        p=f1.find("timer:",p);
        if(p!=f1.npos) { p=std::atoi(f1.substr(p+5).c_str()); if(p < result) result=p; }
    }
    p=f2.find("\"timer\";i:",p);
    if(p!=f2.npos) { p=std::atoi(f2.substr(p+10).c_str()); if(p < result) result=p; }
    
    fprintf(stderr, "= %u */\n", result); fflush(stderr);
    
    return result;    
}
#endif

void BuildFloorCoordinates()
{
    LevelState state;
    state.CreateFromLevel();
    
    std::fill_n(FloorCoordinateTranslation,  MaxDim,      FloorAmount);
    std::fill_n(FloorCoordinateUntranslation,FloorAmount, MaxDim);

    unsigned floor=0;
    for(unsigned pos=0; pos<MaxDim; ++pos)
    {
        if(state.IsWall(pos)) continue;
        
        if(floor >= FloorAmount)
        {
            std::cerr << "Too many floor\n";
            abort();
        }
        
        FloorCoordinateTranslation[pos] = floor;
        FloorCoordinateUntranslation[floor] = pos;
        ++floor;
    }
#if 0
    std::cerr << "FloorCoordinateTrans:\n";
    for(unsigned a=0; a<MaxDim; ++a)
    {
        if(!(a%MaxWidth)) std::cerr << "[" << a << "]";
        std::cerr << FloorCoordinateTranslation[a] << ' ';
        if(!((a+1)%MaxWidth))std::cerr<<"\n";
    }
    
    std::cerr << "FloorCoordinateUntrans:\n";
    for(unsigned a=0; a<FloorAmount; ++a)
    {
        std::cerr << FloorCoordinateUntranslation[a] << ' ';
    }
    std::cerr << "\n";
#endif
}

void BuildCompactCoordinates()
{
    LevelState state;
    state.CreateFromLevel();

    CompactPlayerCoordinateLimit=0;
    CompactBlockCoordinateLimit=1;
    std::fill(CompactPlayerCoordinateTranslation,
              CompactPlayerCoordinateTranslation+MaxDim, 0);
    std::fill(CompactBlockCoordinateTranslation,
              CompactBlockCoordinateTranslation+MaxDim+1, 0);
    for(unsigned testpos=0; testpos<MaxDim; ++testpos)
    {
        if(state.IsWall(testpos)) continue;

        CompactPlayerCoordinateTranslation[testpos] = CompactPlayerCoordinateLimit++;
        if(!DeadZones.CheckDead(testpos))
        {
            CompactBlockCoordinateTranslation[testpos] = CompactBlockCoordinateLimit++;
        }
    }
    
    /* Calculate the number of different possible box configurations */
    BigInt tmp;
    tmp.Choose(CompactBlockCoordinateLimit, ActualMaxObject-1);

    /* Multiply by number of different player positions */
    tmp *= CompactPlayerCoordinateLimit - (ActualMaxObject-1);
    
    const double log_tmp = std::log(tmp.GetDouble());
    
    //static const double log240 = 5.480638923341991;
    static const double log256 = 5.545177444479562;
    const unsigned bytes_required = (unsigned)std::ceil(log_tmp / log256);
    
#if CHOOSE_LEVEL != 60
    std::cout
        << "// Number of player coordinates: "
               << CompactPlayerCoordinateLimit << "  1 (" << FloorAmount << " specified)\n"
           "// Number of block coordinates: "
               << CompactBlockCoordinateLimit << "  " << MaxObject-1 << "\n"
           "// Compact coordinate data requires " << bytes_required
        << " bytes. " << MaxCompactCoordinateData << " given.\n" << std::flush;

    std::cout
        << "//State sizes:\n"
        << "// Uncompressed " << sizeof(LevelStateData)
        << "; Compressed " << sizeof(CompressedLevel)
        << ". Move size is " << sizeof(Move)
        << ", extra offset bits in Move=" << Move::offset_extra_bits
        << "\n" << std::flush;
#endif
    
    if(MaxCompactCoordinateData < bytes_required
    || CompactPlayerCoordinateLimit > FloorAmount)
    {
        abort();
    }

#if CHOOSE_LEVEL == 60
    std::cerr <<
        "static const unsigned MaxCompactCoordinateData=" << bytes_required << ";\n"
        "static const unsigned FloorAmount=" << CompactPlayerCoordinateLimit << ";\n"
        "static const unsigned InitialMaxDuration=" << GetMaxDuration() << ";\n"
        "#endif\n" << std::flush;
#endif
}
