#ifndef bqtDeadZonesHH
#define bqtDeadZonesHH

#include "coordinates.hh"
#include "settings.hh"
#include "mapdata.hh"

#include <bitset>
#include <vector>
#include <map>
#include <iostream>

class BigInt;

/* Dead zones: Areas where blocks must never be moved into. */
class DeadZones
{
public:
    inline bool CheckDead(unsigned pos) const
    {
        return DeadFlags[pos];
    }
    
    bool CheckMultiBlockDead(unsigned playerpos) const;
    bool CheckMultiBlockDead(unsigned playerpos, unsigned lastblockpos) const;
    bool CheckMultiBlockDeadBlock() const;
    bool CheckMultiBlockDeadBlock(unsigned lastblockpos) const;
    
    void BuildFromLevel();
    void CreateDeadFlags1();
    void FindMargins() const;
    
private:
    /*static*/ bool CheckPoint(unsigned x,unsigned y,
                           int wallx,int wally,
                           int dx1,int dy1,
                           int dx2,int dy2) const;
    
    inline char at(unsigned x,unsigned y) const
    {
        return maps[levelno].mapdata[(y+OffsetHeight)*19+(x+OffsetWidth)]&7;
    }
    inline char at(unsigned pos) const
    {
        return at(pos%MaxWidth, pos/MaxWidth);
    }
    
    void Dump() const;

public:
    typedef std::bitset<FloorAmount> BlockSet;
    struct DeadBlkSet
    {
        CoordinateType plrpos;
        BlockSet blockmap;
    };
    typedef std::vector<DeadBlkSet> DeadBlkSetVec;
    typedef std::vector<BlockSet> BlockSetVec;

    template<unsigned Depth>
    void RegisterMultiDead(unsigned plrpos, const unsigned blockposlist[Depth],
                           bool Sort=true);

    template<unsigned Depth>
    void RegisterMultiIgnore(unsigned plrpos, const unsigned blockposlist[Depth],
                             bool Sort=true);
    
    template<unsigned Depth>
    void RegisterAlwaysMultiDead(const unsigned blockposlist[Depth]);

    template<unsigned Depth> void CreateMultiDeadFlags();

private:
    bool DeadFlags[MaxDim];
    
    unsigned UndeadList[MaxDim];
    unsigned UndeadMax;
    
    struct DeadFlagData
    {
        DeadBlkSetVec WithPlayer;
        BlockSetVec Set;
    };
    DeadFlagData MultiDeadFlags;
    DeadFlagData MultiIgnoreFlags;

    std::map<CoordinateType/*normal block coordinate*/, DeadFlagData> MultiDeadFlagsByBlock;
    std::map<CoordinateType/*normal block coordinate*/, DeadFlagData> MultiIgnoreFlagsByBlock;

private:
    void CountCombinations();
    
    double MultiDeadCalculatePosition
       (const class BigInt& total,
        const unsigned char digits[],
        unsigned Depth) const;

    template<unsigned Depth> void LoadDeadData(unsigned char digits[Depth]);
    template<unsigned Depth> void SaveDeadData(const unsigned char digits[Depth]);

    void CreateAllMultiDeadFlags();

     /* returns the number of new dead combinations */
    template<unsigned Depth> void DeadMultiBlockBuilder
       (const unsigned blockposlist[Depth],
        unsigned& ndead,
        unsigned& nalldead,
        unsigned& nignore);

    template<unsigned Depth> void DeadMultiBlockBuilderLoop(unsigned char digits[Depth]);

    bool CheckMultiBlockIgnore(unsigned playerpos, unsigned lastblockpos) const;
    bool CheckMultiBlockIgnore(unsigned playerpos) const;
};
extern DeadZones DeadZones;

#endif
