#include "bigint.hh"

BigInt::BigInt()
{
    mpz_init2(mpz, 64);
}
BigInt::BigInt(unsigned value)
{
    mpz_init(mpz);
    *this = value;
}
BigInt::BigInt(int value)
{
    mpz_init(mpz);
    *this = value;
}
BigInt::BigInt(size_t value)
{
    mpz_init(mpz);
    *this = value;
}
BigInt::BigInt(const BigInt& b)
{
    mpz_init(mpz);
    *this = b;
}
BigInt::~BigInt()
{
    mpz_clear(mpz);
}

BigInt& BigInt::operator= (unsigned v)
{
    mpz_set_ui(mpz, v);
    return *this;
}
BigInt& BigInt::operator= (const BigInt& b)
{
    mpz_set(mpz, b.mpz);
    return *this;
}
BigInt& BigInt::operator<<= (unsigned v)
{
    mpz_mul_2exp(mpz, mpz, v);
    return *this;
}

BigInt& BigInt::operator+= (unsigned v)
{
    mpz_add_ui(mpz, mpz, v);
    return *this;
}
BigInt& BigInt::operator-= (unsigned v)
{
    mpz_sub_ui(mpz, mpz, v);
    return *this;
}
BigInt& BigInt::operator+= (const BigInt& v)
{
    mpz_add(mpz, mpz, v.mpz);
    return *this;
}
BigInt& BigInt::operator-= (const BigInt& v)
{
    mpz_sub(mpz, mpz, v.mpz);
    return *this;
}


BigInt& BigInt::operator*= (unsigned v)
{
    mpz_mul_ui(mpz, mpz, v);
    return *this;
}
BigInt& BigInt::operator/= (unsigned v)
{
    mpz_fdiv_q_ui(mpz, mpz, v);
    return *this;
}

BigInt& BigInt::operator*= (const BigInt& v)
{
    mpz_mul(mpz, mpz, v.mpz);
    return *this;
}

BigInt& BigInt::operator/= (const BigInt& v)
{
    mpz_fdiv_q(mpz, mpz, v.mpz);
    return *this;
}

unsigned BigInt::operator% (unsigned v) const
{
    /*
    static struct tmpf
    {
        mpz_t tmp;
        tmpf() { mpz_init(tmp); }
        ~tmpf() { mpz_clear(tmp); }
    } tmp;
    return mpz_fdiv_r_ui(tmp.tmp, mpz, v);
    */
    return mpz_fdiv_ui(mpz, v);
}

unsigned BigInt::DivMod(BigInt& q, BigInt& r, unsigned v) const
{
    return mpz_fdiv_qr_ui(q.mpz, r.mpz, mpz, v);
}
void BigInt::DivMod(BigInt& q, BigInt& r, const BigInt& v) const
{
    mpz_fdiv_qr(q.mpz, r.mpz, mpz, v.mpz);
}

unsigned BigInt::DivMod2Exp(BigInt& q, BigInt& r, unsigned v) const
{
    mpz_fdiv_r_2exp(r.mpz, mpz, v);
    mpz_fdiv_q_2exp(q.mpz, mpz, v);
    return mpz_get_ui(r.mpz);
}

void BigInt::Pow(unsigned b,unsigned e)
{
    mpz_ui_pow_ui(mpz, b, e);
}

void BigInt::Choose(unsigned n,unsigned k)
{
    if(!k) { *this=1; return; }
    if(n<k) { *this=0; return; }
    mpz_bin_uiui(mpz, n,k);
}

BigInt::operator bool() const
{
    return mpz_sgn(mpz) != 0;
}

size_t BigInt::GetInt() const
{
    return mpz_get_ui(mpz);
}

double BigInt::GetDouble() const
{
    return mpz_get_d(mpz);
}

unsigned BigInt::GetLog2Ceil() const
{
    signed long int exp;
    /*double tmp = */mpz_get_d_2exp(&exp, mpz);
    
    /* 0.5 <= tmp < 1.0 */
    /* As such, it can be ignored. */
    return exp;
}
