#include "simplevec.hh"
#include "settings.hh" /* For MaxObject, MaxDim */
#include "binom.hh"

namespace Binom
{
    struct CacheLine
    {
    public:
        inline BigInt& operator[] (unsigned n)
        {
            if(unlikely(n >= MaxObject))
                { fprintf(stderr, "Choose Error: n=%u, MaxObject=%u\n", n,MaxObject); throw -1;} 
            return line[n];
        }
        void InitializeIfNeeded(unsigned n)
        {
            if(likely(line)) return;
            line = new BigInt[MaxObject];
            for(unsigned k=1; k<=MaxObject; ++k)
                line[k-1].Choose(n,k);
        }
        CacheLine(): line(0) {}
        ~CacheLine() { delete[] line; }
    private:
        BigInt* line;
    };
    const BigInt& Choose(unsigned n, unsigned k)
    {
        static struct Data
        {
            BigInt ichi, zero;
            SimpleVec<CacheLine, MaxDim> binomcache;
            Data(): ichi(1), zero(0) { }
            CacheLine& operator[] (unsigned n)
            {
                if(unlikely(n > (unsigned)MaxDim))
                    { fprintf(stderr, "Choose Error: n=%u, MaxDim=%u\n", n, MaxDim); throw -1; }
                
                CacheLine& line = binomcache[n];
                line.InitializeIfNeeded(n);
                return line;
            }
        } Data;
        
        if(k==0) return Data.ichi;
        if(unlikely(n<k)) return Data.zero;
        return Data[n][k-1];
    }
}
