#ifndef bqtNewRadixMapHH
#define bqtNewRadixMapHH

#include "RadixTreeGMP-Vector.hh"
#include "NewRadixUtils.hh"

template<unsigned RealSize, typename DataType>
class NewRadixMapGMP: private RadixTreeGMP<RealSize,sizeof(DataType)>
{
private:
    typedef RadixTreeGMP<RealSize,sizeof(DataType)> parent;
public:
    using parent::size;
    using parent::clear;
    inline bool find(BigInt& Key, DataType& value) const
    {
        return parent::find(Key, (unsigned char*)&value);
    }
    inline bool add(BigInt& Key, const DataType& value)
    {
        return parent::add(Key, (const unsigned char*)&value);
    }
    inline bool add_or_find(BigInt& Key, DataType& value)
    {
        return parent::add_or_find(Key, (unsigned char*)&value);
    }
    inline void ensure(BigInt& Key, const DataType& value)
    {
        parent::ensure(Key, (const unsigned char*)&value);
    }
};

/* Note: BytesUsedBy won't work if MaxValue-MinValue == (unsigned)~0 */
template<typename KeyType, unsigned MaxValue, unsigned KeyLength, typename DataType, unsigned MinValue=0>
class NewRadixMap:
    private NewRadixMapGMP<BytesUsedBy<MaxValue-MinValue+1,KeyLength>::MaxBytes, DataType>
{
private:
    typedef NewRadixMapGMP<BytesUsedBy<MaxValue-MinValue+1,KeyLength>::MaxBytes, DataType> parent;
    static inline BigInt& Encode(const KeyType*const Key)
    {
        return NewRadixEncoder::Encode<KeyType,MaxValue,MinValue> (Key, KeyLength);
    }
public:
    using parent::size;
    using parent::clear;
    inline bool find(const KeyType*const Key, DataType& value) const
    {
        return parent::find(Encode(Key), value);
    }
    inline bool add(const KeyType*const Key, const DataType& value)
    {
        return parent::add(Encode(Key), value);
    }
    inline bool add_or_find(const KeyType*const Key, DataType& value)
    {
        return parent::add_or_find(Encode(Key), value);
    }
    inline void ensure(const KeyType*const Key, const DataType& value)
    {
        parent::ensure(Encode(Key), value);
    }
};

#endif
