#include "filewindow.hh"

#include "moves.hh"

static class MoveContainer
{
private:
    static const unsigned WindowSize = 0x10000;
    static const unsigned NumWindows = 2;
    
    typedef TempFileBuffer<WindowSize, NumWindows> NodeType;

    NodeType Node;
    off64_t LastPtr;
    
public:
    MoveContainer(): Node(), LastPtr(0), nmoves(0),n1(0),n2(0),n3(0),n4(0)
    {
    }
    
    const Move& ReadMove(off64_t ptr) const
    {
        unsigned char* data = Node.GetWindow(ptr, sizeof(Move)+4);
        return *(Move*)data;
    }
    off64_t ReadMovePrevPtr(off64_t ptr) const
    {
        unsigned char* data = Node.GetWindow(ptr, sizeof(Move)+4);
        
        const Move& mv = *(Move*)data;
        unsigned bytes = mv.offsetbytes+1;
        off64_t result = 0;
        for(unsigned a=0; a<bytes; ++a)
        {
            result |= data[sizeof(Move)+a] << (a*8);
        }
        result <<= Move::offset_extra_bits;
        result |= mv.offset_extra;
        return ptr-result;
    }
    off64_t AddMove(Move mv, off64_t prevptr)
    {
        ++nmoves;
        off64_t address = LastPtr;
        off64_t offset  = address - prevptr;
        
        avg_link += offset; avg_count += 1;
        
        mv.offset_extra = offset % (1 << Move::offset_extra_bits);
        offset >>= Move::offset_extra_bits;
        
        unsigned bytes;
        if(offset & 0xFFFF0000)
            bytes = (offset & 0xFF000000) ? 4 : 3;
        else
            bytes = (offset & 0xFF00) ? 2 : 1;
        
        mv.offsetbytes  = bytes-1;
        
        unsigned size = sizeof(mv)+bytes;
        
        unsigned char* newbytes = Node.GetWindow(address, size);

        std::memcpy(newbytes, (void*)&mv, sizeof(Move));
        
        for(unsigned a=0; a<bytes; ++a)
        {
            newbytes[sizeof(Move)+a] = (offset >> (a*8)) & 0xFF;
        }
        
        switch(bytes)
        {
            case 1: ++n1; break;
            case 2: ++n2; break;
            case 3: ++n3; break;
            case 4: ++n4; break;
        }
        
        LastPtr = address+size;
        
        return address;
    }
    void DelMove(off64_t ptr)
    {
    }
    void Report(std::ostream& out) const
    {
        double avg = nmoves ? (LastPtr/(double)nmoves) : 0.0;
        unsigned avg_l = avg_count ? avg_link/avg_count : 0;
        
        out << "// MoveContainer has " << nmoves
                   << " (" << LastPtr << " bytes, "
                   << avg << " average"
                   << " (" << n1 << "," << n2 << "," << n3 << "," << n4 << ")"
                   << " avg_link(" << avg_l << ")"
                   << ")\n";
        avg_link=avg_count=0;
    }
    mutable off64_t avg_link, avg_count;
private:
    unsigned nmoves;
    unsigned n1,n2,n3,n4;
} MoveContainer;

void MoveTree::push_back(const Move& t)
{
    Last = MoveContainer.AddMove(t, Last);
}
void MoveTree::clear()
{
    if(empty()) return;
    if(Last != Root) MoveContainer.DelMove(Last);
    Last=Root;
}

MoveTree::reverse_const_iterator& MoveTree::reverse_const_iterator::operator++ ()
{
    ptr = MoveContainer.ReadMovePrevPtr(ptr);
    return *this;
}
const Move& MoveTree::reverse_const_iterator::operator* ()
{
    return MoveContainer.ReadMove(ptr);
}

void Move::Dump(std::ostream& out) const
{
    static const char* const tt[2] = {"Kick ","Push "};
    static const char* const dt[4] = {"Up","Left","Right","Down"};
#ifdef TIMESTAMP_MOVES
    out << "[" << timestamp << "]";
#endif
    out << tt[type] << dt[dir]
        << " @ "
        << getpx() << ',' << getpy();
}
