//#define PAIR_ATTRIBUTES __attribute__((packed))

#ifndef __STDC_CONSTANT_MACROS
#define __STDC_CONSTANT_MACROS /* for UINT16_C etc */
#endif

#define _LARGEFILE64_SOURCE
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>

#include <algorithm>
#include <climits>
#include <iostream>
#include <sstream>

#include <csignal>

#ifdef __CYGWIN__
# define off64_t _off64_t
# define O_LARGEFILE 0
# define lseek64 lseek
# define mmap64 mmap
#endif

#include "levelstate.hh"
#include "deadzones.hh"
#include "completer.hh"

#if CHOOSE_LEVEL == 60
# include <sys/wait.h>
#endif

static void CompleteLevel()
{
    DeadZones.BuildFromLevel();
    
    if(TerminateSignalled)
    {
        std::cerr << "Terminate signalled, not solving a level!\n";
        return;
    }

    LevelState state;
    state.BuildFromLevel();
    state.RebuildBlockCaches();
    
    Completer completer(state);
    
    PlayingNormally=true;
//    ShowDead=true;    
    while(!TerminateSignalled)
    {
        bool Finished = completer.Iteration();
        if(Finished) break;
    }
    PlayingNormally=false;
    completer.DumpBest();
    
    std::cout << completer.GetJs();
}

static void CompleteLevel(int phase,int level)
{
    levelno = (phase-1)*5+level-1;
    std::cout << "/* LEVEL " << levelno << " */\n";
    CompleteLevel();
}

static void Terminate(int sig)
{
    TerminateSignalled=true;
}

Completer::active_list_type Completer::active_list;

int main(int argc, const char* const* argv)
{
    signal(SIGTERM, Terminate);
    signal(SIGINT,  Terminate);
    
#if CHOOSE_LEVEL == 60
    std::cerr <<
        "typedef " << CompactCoordinateUnitType << " CompactCoordinateUnit;\n"
        "enum { CompactCoordinateBitsPerUnit = " << CompactCoordinateBitsPerUnit << " };\n"
        "\n";

    static std::map<int, unsigned> pid_list;
    const size_t NUM_FORKS = 1;
    for(levelno=0; levelno<60; ++levelno)
    {
        static int pid;
        {int waitflag = (pid_list.size() < NUM_FORKS) ? WNOHANG : 0;
        do {
            int status=0;
            int pid = waitpid(-1, &status, waitflag);
            if(pid > 0)
            {
                if(WIFSIGNALED(status))
                {
                    fprintf(stderr, "Child died at signal %d\n",
                        WTERMSIG(status));
                }  
                pid_list.erase(pid);
            }
            waitflag=0;
        } while(pid_list.size() >= NUM_FORKS); }
        
        std::cout << std::flush; fflush(stdout);
        std::cerr << std::flush; fflush(stderr);
        pid = fork();
        if(pid > 0)
        {
            pid_list[pid] = levelno;
            //fprintf(stderr, "."); fflush(stderr);
            continue;
        }
        if(pid < 0) perror("fork");
        
        DeadZones.BuildFromLevel();
        std::cerr << std::flush;
        std::cout << std::flush;
        fflush(stdout);
        fflush(stderr);
        _exit(0);
    }
    while(pid_list.size() > 0)
    {
        int status=0;
        int pid = waitpid(-1, &status, 0);
        if(pid > 0)
        {
            if(WIFSIGNALED(status))
            {
                fprintf(stderr, "Child died at signal %d\n",
                    WTERMSIG(status));
            }
            pid_list.erase(pid);
        }
    }
    return 0;
#endif
#if 0
    LevelState state;
    state.BuildFromLevel();
    state.RebuildBlockCaches();
    state.Walk(-1,-2);
    state.Kick(-1,0,1);
    state.Walk(0,-2);
    state.CanPush(-1);
    state.Kick(-1,0,1);
    state.Walk(-3,3);
    //state.CanPush(0,-1);
    return 0;
#endif
    
    if(argc < 3)
    {
        levelno=CHOOSE_LEVEL;
        CompleteLevel();
        return 0;
    }
    
    CompleteLevel(std::atoi(argv[1]), std::atoi(argv[2]));
    return 0;
}
