#include "coordinates.hh"

#include "levelstate.hh"
#include "deadzones.hh"
#include "moves.hh"

#include <algorithm>
#include <iostream>
#if CHOOSE_LEVEL == 60
# include <sstream>
#endif
#include <cmath> /* For std::log */

volatile bool TerminateSignalled = false;
size_t levelno = CHOOSE_LEVEL;
size_t MaxDuration = InitialMaxDuration;
size_t ActualMaxObject=0;

size_t CompactPlayerCoordinateTranslation[MaxDim];
size_t CompactPlayerCoordinateLimit;
size_t CompactBlockCoordinateTranslation[MaxDim+1];
size_t CompactBlockCoordinateLimit;
size_t FloorCoordinateTranslation[MaxDim];
size_t FloorCoordinateUntranslation[FloorAmount];

#if CHOOSE_LEVEL == 60
static size_t GetMaxDuration()
{
    size_t result = 60*30*10; // 10 minutes
    
    std::string f1,f2;
    
    {std::stringstream s;
    s << "../solutions/" << levelno << ".txt";
    f1 = s.str();}
    
    {std::stringstream s;
    s << "../records/" << levelno << ".txt";
    f2 = s.str();}
    
    fprintf(stderr, "/* completion time of %s and %s ",
        f1.c_str(),f2.c_str()); fflush(stderr);
    
    {FILE*fp=fopen(f1.c_str(),"rb");if(!fp)f1="";else{fseek(fp,0,SEEK_END);
     std::vector<unsigned char> buf(ftell(fp));rewind(fp);
     fread(&buf[0],1,buf.size(),fp);fclose(fp); f1.assign(buf.begin(),buf.end()); }}

    {FILE*fp=fopen(f2.c_str(),"rb");if(!fp)f2="";else{fseek(fp,0,SEEK_END);
     std::vector<unsigned char> buf(ftell(fp));rewind(fp);
     fread(&buf[0],1,buf.size(),fp);fclose(fp); f2.assign(buf.begin(),buf.end()); }}
    
    //fprintf(stderr, "/*%u,%u*/", (unsigned)f1.size(), (unsigned)f2.size());
    
    std::string::size_type p = f1.find("perfect_data=");
    //fprintf(stderr, "/*%lu", (unsigned long)p);
    if(p!=f1.npos)
    {
        p=f1.find("timer:",p);
        if(p!=f1.npos) { p=std::atoi(f1.c_str()+p+6); if(p < result) result=p; }
    }
    p=f2.find("\"timer\";i:");
    //fprintf(stderr, ",%lu*/", (unsigned long)p);
    if(p!=f2.npos) { p=std::atoi(f2.c_str()+p+10); if(p < result) result=p; }
    
    fprintf(stderr, "= %lu */\n", (unsigned long)result); fflush(stderr);
    
    return result;    
}
#endif

void BuildFloorCoordinates()
{
    LevelState state;
    state.BuildFromLevel();
    
    std::fill_n(FloorCoordinateTranslation,  (size_t)MaxDim,      (size_t)FloorAmount);
    std::fill_n(FloorCoordinateUntranslation,(size_t)FloorAmount, (size_t)MaxDim);

    size_t floor=0;
    for(size_t pos=0; pos< (size_t)MaxDim; ++pos)
    {
        if(state.IsWall(pos)) continue;
        
        if(floor >= FloorAmount)
        {
            std::cerr << "Too many floor\n";
            abort();
        }
        
        FloorCoordinateTranslation[pos] = floor;
        FloorCoordinateUntranslation[floor] = pos;
        ++floor;
    }
#if 0
    std::cerr << "FloorCoordinateTrans:\n";
    for(size_t a=0; a< (size_t)MaxDim; ++a)
    {
        if(!(a%MaxWidth)) std::cerr << "[" << a << "]";
        std::cerr << FloorCoordinateTranslation[a] << ' ';
        if(!((a+1)%MaxWidth))std::cerr<<"\n";
    }
    
    std::cerr << "FloorCoordinateUntrans:\n";
    for(size_t a=0; a<FloorAmount; ++a)
    {
        std::cerr << FloorCoordinateUntranslation[a] << ' ';
    }
    std::cerr << "\n";
#endif
}

void BuildCompactCoordinates()
{
    LevelState state;
    state.BuildFromLevel();

    CompactPlayerCoordinateLimit=0;
    CompactBlockCoordinateLimit=1; // 0 = no block
    std::fill(CompactPlayerCoordinateTranslation,
              CompactPlayerCoordinateTranslation+MaxDim, 0);
    std::fill(CompactBlockCoordinateTranslation,
              CompactBlockCoordinateTranslation+MaxDim+1, 0);
    for(size_t testpos=0; testpos< (size_t)MaxDim; ++testpos)
    {
        if(state.IsWall(testpos)) continue;

        CompactPlayerCoordinateTranslation[testpos] = CompactPlayerCoordinateLimit++;
        if(!DeadZones.CheckDead(testpos))
        {
            CompactBlockCoordinateTranslation[testpos] = CompactBlockCoordinateLimit++;
        }
    }
    
    /* Calculate the number of different possible box configurations */
    BigInt tmp;
    tmp.Choose(CompactBlockCoordinateLimit, ActualMaxObject-1);

    /* Multiply by number of different player positions */
    tmp *= CompactPlayerCoordinateLimit - (ActualMaxObject-1);
    
    const double log_tmp      = std::log(tmp.GetDouble());
    const double log_tmp_div8 = std::log(tmp.GetDouble() / 8.0);
    
    const double log_compact = std::log(1 << CompactCoordinateBitsPerUnit);
    
    //static const double log240 = 5.480638923341991;
    static const double log256 = 5.545177444479562;
    const size_t bytes_required = (size_t)std::ceil(log_tmp / log_compact);

    const size_t bytes_required_div8 = (size_t)std::ceil(log_tmp_div8 / log256);
    
#if CHOOSE_LEVEL != 60
    std::cout
        << "// Number of player coordinates: "
               << CompactPlayerCoordinateLimit << "  1 (" << FloorAmount << " specified)\n"
           "// Number of block coordinates: "
               << CompactBlockCoordinateLimit << "  " << MaxObject-1 << "\n"
           "// Compact coordinate data requires " << bytes_required
        << " " << CompactCoordinateBitsPerUnit
        << "-bit units. " << MaxCompactCoordinateData << " given.\n" << std::flush;

    std::cout
        << "//State sizes:\n"
        << "// Uncompressed " << sizeof(LevelState)
        << "; Compressed " << sizeof(CompressedLevel)
        << ". Move size is " << sizeof(Move)
        << ", extra offset bits in Move=" << Move::offset_extra_bits
        << "\n" << std::flush;
#endif
    
    if(MaxCompactCoordinateData < bytes_required
    || CompactPlayerCoordinateLimit > FloorAmount)
    {
        abort();
    }

#if CHOOSE_LEVEL == 60
    std::cerr <<
        "enum {  MaxCompactCoordinateData=" << bytes_required << " };\n"
        "enum {  MaxBitsetCoordinateData=" << bytes_required_div8 << " };\n"
        "enum {  FloorAmount=" << CompactPlayerCoordinateLimit << " };\n"
        "//enum {  BlockCoordAmount=" << CompactBlockCoordinateLimit << " };\n"
        "enum {  InitialMaxDuration=" << GetMaxDuration() << " };\n"
        "#endif\n" << std::flush;
#endif
}
