#include "NewRadixSet.hh"

class DeadCompleter
{
private:
    enum VerdictType { Unknown=0, Ok=1, Dead=2 };
public:
    DeadCompleter(const LevelState& lev)
    {
        ReinitWith(lev);
    }
    
    void ReinitWith(const LevelState& lev)
    {
        initial_state = lev;
        /* Don't clear the seenmap. */
    }
    
    void Process()
    {
        Process(initial_state);
    }
    
private:
    inline void CombineResult(VerdictType& result, VerdictType state)
    {
        switch(state)
        {
            case Unknown: if(result==Dead) result=Unknown; break;
            case Ok:      result=Ok; break;
            case Dead:    /* keep the existing state */ break;
        }
    }

    VerdictType TryMoves(const LevelState& state)
    {
        state.RebuildBlockCaches();
        if(state.IsCompleted())
        {
            /* Not a blunder, because it's completed! */
            /* And don't try more moves. */
            return Ok;
        }

        typedef LevelState::FavorablePointList targetlist;
        
        targetlist targets;
        state.FindWalkingDestinations(targets);
        
        VerdictType result = Dead;
        for(targetlist::const_iterator i = targets.begin(); i != targets.end(); ++i)
        {
            const CoordinateType Point = i->Point;
            
            if(i->Dir & 4) CombineResult(result, TryBranch(state, Point, -1,      ));
            if(i->Dir & 8) CombineResult(result, TryBranch(state, Point,  1,      ));
            if(i->Dir & 1) CombineResult(result, TryBranch(state, Point, -MaxWidth));
            if(i->Dir & 2) CombineResult(result, TryBranch(state, Point,  MaxWidth));
        }
        if(result == Dead)
        {
            unsigned plrpos = state.GetPlayerPos();
            unsigned blockposlist[Depth];
            for(unsigned a=0; a<Depth; ++a)
                blockposlist[a] = state.GetBlockPos(a);

            state.Dump();
            DeadZones.RegisterMultiDead<Depth> (plrpos, blockposlist);
        }
        return result;
    }
    
    VerdictType
        TryBranch(const LevelState& refstate,
                  const CoordinateType Point, int offset)
    {
        if(!refstate.CanPushFrom(Point, offset))
        {
            return Dead;
        }
        
        
        LevelState NewState = refstate;
        
        NewState.PushFromNoMove(Point, offset);
        
        if(NewState.CheckBlunder(Point+offset*2,offset))
        {
            refstate.RebuildBlockCaches();
            return Dead;
        }
        
        NewState.WalkTo(Point+offset);
        
        VerdictType result = Unknown;
        if(seenmap.add_or_find(NewState.GetData(), result))
        {
            // Subprocess this state
            result = TryMoves(NewState);
            if(result != Unknown)
                seenmap.ensure(NewState.GetData(), result);
        }
        return result;
    }

private:
    LevelState initial_state;
    
    typedef NewRadixMap<CoordinateType,MaxDim,MaxObject, VerdictType> seenmaptype;
    seenmaptype seenmap;
};
