#ifndef bqtDeadZonesHH
#define bqtDeadZonesHH

#include "coordinates.hh"
#include "settings.hh"
#include "mapdata.hh"

#include <bitset>
#include <vector>
#include <map>
#include <iostream>

class BigInt;

/* Dead zones: Areas where blocks must never be moved into. */
class DeadZones
{
public:
    inline bool CheckDead(size_t pos) const
    {
        return DeadFlags[pos] > 1;
    }
    inline bool CheckUnmovable(size_t pos) const
    {
        return DeadFlags[pos];
    }
    
    bool CheckMultiBlockDead(size_t playerpos) const;
    bool CheckMultiBlockDead(size_t playerpos, size_t lastblockpos) const;
    bool CheckMultiBlockDeadBlock() const;
    bool CheckMultiBlockDeadBlock(size_t lastblockpos) const;
    
    void BuildFromLevel();
    void CreateDeadFlags1();
    void FindMargins() const;
    
private:
    static int CheckPoint(size_t point, int walldir, int sidedir) REGPARM;

public:
    typedef std::bitset<FloorAmount> BlockSet;
    struct DeadBlkSet
    {
        CoordinateType plrpos;
        BlockSet blockmap;
    };
    typedef std::vector<DeadBlkSet> DeadBlkSetVec;
    typedef std::vector<BlockSet> BlockSetVec;

    template<size_t Depth>
    void RegisterMultiDead(size_t plrpos, const ssize_t blockposlist[Depth],
                           bool Sort=true);

    template<size_t Depth>
    void RegisterMultiIgnore(size_t plrpos, const ssize_t blockposlist[Depth],
                             bool Sort=true);
    
    template<size_t Depth>
    void RegisterAlwaysMultiDead(const ssize_t blockposlist[Depth]);

    template<size_t Depth> void CreateMultiDeadFlags();

private:
    char DeadFlags[MaxDim]; // 0=not, 1=unmovable, 2+ = unmovable and dead
    
    ssize_t UndeadList[MaxDim];
    size_t UndeadMax;
    
    struct DeadFlagData
    {
        DeadBlkSetVec WithPlayer;
        BlockSetVec Set;
    };
    DeadFlagData MultiDeadFlags;
    DeadFlagData MultiIgnoreFlags;

    std::map<CoordinateType/*normal block coordinate*/, DeadFlagData> MultiDeadFlagsByBlock;
    std::map<CoordinateType/*normal block coordinate*/, DeadFlagData> MultiIgnoreFlagsByBlock;

private:
    void CountCombinations();
    
    double MultiDeadCalculatePosition
       (const class BigInt& total,
        const unsigned char digits[],
        size_t Depth) const __attribute__((noinline));

    template<size_t Depth> void LoadDeadData(unsigned char digits[Depth]);
    template<size_t Depth> void SaveDeadData(const unsigned char digits[Depth]);

    void CreateAllMultiDeadFlags();

     /* returns the number of new dead combinations */
    template<size_t Depth> void DeadMultiBlockBuilder
       (const ssize_t blockposlist[Depth],
        size_t& ndead,
        size_t& nalldead,
        size_t& nignore);

    template<size_t Depth> void DeadMultiBlockBuilderLoop(unsigned char digits[Depth]);

    bool CheckMultiBlockIgnore(size_t playerpos, size_t lastblockpos) const;
    bool CheckMultiBlockIgnore(size_t playerpos) const;
};
extern DeadZones DeadZones;

#endif
