#include "endian.hh"
#include <algorithm>

#include <stdlib.h> // size_t
#include "hash64.hh"

/* From http://burtleburtle.net/bob/c/lookup8.c */

#define mix64(a,b,c) \
{ \
  a -= b; a -= c; a ^= (c>>43); \
  b -= c; b -= a; b ^= (a<<9); \
  c -= a; c -= b; c ^= (b>>8); \
  a -= b; a -= c; a ^= (c>>38); \
  b -= c; b -= a; b ^= (a<<23); \
  c -= a; c -= b; c ^= (b>>5); \
  a -= b; a -= c; a ^= (c>>35); \
  b -= c; b -= a; b ^= (a<<49); \
  c -= a; c -= b; c ^= (b>>11); \
  a -= b; a -= c; a ^= (c>>12); \
  b -= c; b -= a; b ^= (a<<18); \
  c -= a; c -= b; c ^= (b>>22); \
}

typedef uint_least64_t ub8;

uint_fast64_t hash64(const unsigned char*k, size_t length, uint_fast64_t level)
{
  register ub8 a,b,c,len;

  /* Set up the internal state */
  len = length;
  a = b = level;                         /* the previous hash value */
  c = UINT64_C(0x9e3779b97f4a7c13); // 2^64 / ((1+sqrt(5))/2)

  /*---------------------------------------- handle most of the key */
  while (len >= 24)    /* aligned */
  {
    a += R64(k+0);
    b += R64(k+8);
    c += R64(k+16);
    mix64(a,b,c);
    k += 24; len -= 24;
  }

  /*------------------------------------- handle the last 23 bytes */
  c += length;
  if(len >16)      { c += (ub8)Rn(k+15, len-15) & ~UINT64_C(0xFF);
                     b += (ub8)R64(k+8); 
                     a += (ub8)R64(k); }
  else if(len > 8) { b += (ub8)Rn(k+8,  std::min(8UL, len-8));
                     a += (ub8)R64(k); }
  else               a += (ub8)Rn(k, std::min(8UL, len));
  mix64(a,b,c);
  /*-------------------------------------------- report the result */
  return c;
}
