#include "filewindow.hh"

#include "templates.hh"
#include "moves.hh"

#if 0

static class MoveContainer
{
private:
    static const size_t WindowSize = 0x10000;
    static const size_t NumWindows = 2;
    
    typedef TempFileBuffer<WindowSize, NumWindows> NodeType;

    NodeType Node;
    size_t LastPtr;
    
public:
    MoveContainer(): Node(), LastPtr(0), nmoves(0)
    {
        for(unsigned a=0; a<8; ++a) linkcounts[a] = 0;
    }
    
    const Move& ReadMove(size_t ptr) const
    {
        unsigned char* data = Node.GetWindow(ptr, sizeof(Move)+4);
        return *(Move*)data;
    }
    size_t ReadMovePrevPtr(size_t ptr) const
    {
        unsigned char* data = Node.GetWindow(ptr, sizeof(Move)+4);
        
        const Move& mv = *(Move*)data;
        size_t bytes = mv.offsetbytes+1;
        size_t result = 0;
        for(size_t a=0; a<bytes; ++a)
        {
            result |= data[sizeof(Move)+a] << (a*8);
        }
        result <<= Move::offset_extra_bits;
        result |= mv.offset_extra;
        return ptr-result;
    }
    size_t AddMove(Move mv, size_t prevptr)
    {
        ++nmoves;
        size_t address = LastPtr;
        size_t offset  = address - prevptr;
        
        avg_link += offset; avg_count += 1;
        
        mv.offset_extra = offset % (1 << Move::offset_extra_bits);
        offset >>= Move::offset_extra_bits;
        
        size_t bytes = CalcBytesNeededByInt(offset);
        
        mv.offsetbytes  = bytes-1;
        
        size_t size = sizeof(mv)+bytes;
        
        unsigned char* newbytes = Node.GetWindow(address, size);

        std::memcpy(newbytes, (void*)&mv, sizeof(Move));
        
        for(size_t a=0; a<bytes; ++a)
        {
            newbytes[sizeof(Move)+a] = (offset >> (a*8)) & 0xFF;
        }
        
        ++linkcounts[bytes-1];
        
        LastPtr = address+size;
        
        return address;
    }
    void DelMove(size_t ptr)
    {
    }
    void Report(std::ostream& out) const
    {
        double avg = nmoves ? (LastPtr/(double)nmoves) : 0.0;
        size_t avg_l = avg_count ? avg_link/avg_count : 0;
        
        out << "// MoveContainer has " << nmoves
                   << " (" << LastPtr << " bytes, "
                   << avg << " average"
                   << " (" << linkcounts[0]
                    << "," << linkcounts[1]
                    << "," << linkcounts[2]
                    << "," << linkcounts[3]
                    << "," << linkcounts[4]
                    << "," << linkcounts[5]
                    << "," << linkcounts[6]
                    << "," << linkcounts[7]
                    << ")"
                   << " avg_link(" << avg_l << ")"
                   << ")\n";
        avg_link=avg_count=0;
    }
    mutable size_t avg_link, avg_count;
private:
    size_t nmoves;
    size_t linkcounts[8];
} MoveContainer;

#else

#include <deque>
#include <utility>
static class MoveContainer
{
private:
    typedef std::pair<Move, size_t> movepair;
    std::deque<movepair> moves;
public:
    MoveContainer(): moves() { }

    const Move& ReadMove(size_t ptr) const
    {
        return moves[ptr].first;
    }
    
    size_t ReadMovePrevPtr(size_t ptr) const
    {
        return moves[ptr].second;
    }
    size_t AddMove(Move mv, size_t prevptr)
    {
        size_t res = moves.size();
        moves.push_back(movepair(mv, prevptr));
        return res;
    }
    void DelMove(size_t ptr)
    {
    }
} MoveContainer;

#endif


void MoveTree::push_back(const Move& t)
{
    Last = MoveContainer.AddMove(t, Last);
}
void MoveTree::clear()
{
    if(empty()) return;
    if(Last != Root) MoveContainer.DelMove(Last);
    Last=Root;
}

MoveTree::reverse_const_iterator& MoveTree::reverse_const_iterator::operator++ ()
{
    ptr = MoveContainer.ReadMovePrevPtr(ptr);
    return *this;
}
const Move& MoveTree::reverse_const_iterator::operator* ()
{
    return MoveContainer.ReadMove(ptr);
}

void Move::Dump(std::ostream& out) const
{
    static const char* const tt[2] = {"Kick ","Push "};
    static const char* const dt[4] = {"Up","Left","Right","Down"};
#ifdef TIMESTAMP_MOVES
    out << "[" << timestamp << "]";
#endif
    out << tt[type] << dt[dir]
        << " @ "
        << getpx() << ',' << getpy();
}
